
from typing import Tuple

import torch, math
from torch import Tensor
from torch.distributed._tensor import DTensor
from torch.optim import Optimizer

from .quantizer import (
    BF16,
    LinearOptimState2bit, LinearOptimState3bit, LinearOptimState4bit, LinearOptimState8bit,
    DEOptimState2bit, DEOptimState3bit, DEOptimState4bit, DEOptimState8bit,
    LinearSROptimState2bit, LinearSROptimState3bit, LinearSROptimState4bit, LinearSROptimState8bit,
    DESROptimState2bit, DESROptimState3bit, DESROptimState4bit, DESROptimState8bit,
    QemaOptimState2bit, QemaOptimState3bit, QemaOptimState4bit, QemaOptimState8bit
)
from .quantizer.utils import init_adaq_generator


QUANTIZERS = {
    2: {
        'none': None, 
        'linear': LinearOptimState2bit, 'de': DEOptimState2bit, 
        'linear-sr': LinearSROptimState2bit, 'de-sr': DESROptimState2bit,
        'qema': QemaOptimState2bit
    },
    3: {
        'none': None, 
        'linear': LinearOptimState3bit, 'de': DEOptimState3bit, 
        'linear-sr': LinearSROptimState3bit, 'de-sr': DESROptimState3bit,
        'qema': QemaOptimState3bit
    },
    4: {
        'none': None, 
        'linear': LinearOptimState4bit, 'de': DEOptimState4bit, 
        'linear-sr': LinearSROptimState4bit, 'de-sr': DESROptimState4bit,
        'qema': QemaOptimState4bit
    },
    8: {
        'none': None, 
        'linear': LinearOptimState8bit, 'de': DEOptimState8bit, 
        'linear-sr': LinearSROptimState8bit, 'de-sr': DESROptimState8bit,
        'qema': QemaOptimState8bit
    },
    16: {
        'none': None, 
        'bf16': BF16
    },
}

PACKABLE = {
    2: 4,
    3: 2,
    4: 2,
    8: 1,
    16: 1,
}

class AdamWQ(Optimizer):

    r"""
    AdamW with Quantized states.

    Parameters:
    -----------
    bits: Tuple[int, int]
        (bits for 1st state, bits for 2nd state)
    quantile: float
        quantile for 2nd state logarithmic quantization (qema)
    block_sizes: Tuple[int, int]
        (block size for 1st state, block size for 2nd state)
        Then the block-wise quantization will be performed.
        See (Dettmers T., et al. 8-bit optimizers via block-wise quantization. ICLR, 2022.)[http://arxiv.org/abs/2110.02861] for details.
    quantizers: Tuple[str, str]
        (quantizer for 1st state, quantizer for 2nd state)
        -`none`: The orginal 32-bit state
        -`bf16`: The BF16 format
        -`de`: The dynamic exponent mapping without a stochastic rounding
        -`de-sr`: The dynamic exponent mapping with a stochastic rounding
        -`linear`: The linear mapping without a stochastic rounding
        -`linear-sr`: The linear mapping with a stochastic rounding
        -`qema`: The proposed logarithmic quantization
    min_quantizable_tensor_size: int
        A tensor whose size is less than `min_quantizable_tensor_size` will be excluded from quantization.

    Examples:
    ---------
    >>> model: torch.nn.Module
    >>> from solo.adamw import AdamWQ
    >>> optimizer = AdamWQ(
        model.parameters(),
        lr = 0.001,
        weight_decay = 0.,
        betas = (0.8, 0.999),
        bits = (4, 2),
        quantile = 0.1,
        block_sizes = (128, 128),
        quantizers = ('de', 'qema'),
        min_quantizable_tensor_size = 128
    )
    """

    def __init__(
        self, 
        params, 
        lr: float = 0.001, 
        weight_decay: float = 0.01, 
        betas: Tuple[float] = (0.9, 0.999),
        eps: float = 1e-8, 
        *, 
        bits: Tuple[int] = (4, 2),
        quantile: float = 0.1,
        block_sizes: Tuple[int] = (128, 128),
        quantizers: Tuple[str] = ('de', 'qema'),
        min_quantizable_tensor_size: int = 0,
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        defaults = dict(
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            quantile=quantile
        )

        super().__init__(params, defaults)
        self.bits = bits
        self.quantizers = quantizers
        self.block_sizes = (int(block_sizes[0]), int(block_sizes[1]))
        self.packable_sizes = (PACKABLE[bits[0]], PACKABLE[bits[1]])
        self.min_quantizable_tensor_size = min_quantizable_tensor_size

        init_adaq_generator()

        print(self)

    def __str__(self):
        return f"AdamWQ: [{self.quantizers[0]}-{self.quantizers[1]}]" \
                f"[{self.bits[0]}-{self.bits[1]}]" \
                f"[{self.block_sizes[0]}-{self.block_sizes[1]}]"

    def _init_state(self, p: Tensor, signed: bool, quantile: float):
        local_p = p.to_local() if isinstance(p, DTensor) else p

        state_idx = 0 if signed else 1
        block_size = self.block_sizes[state_idx]
        packable_size = self.packable_sizes[state_idx]
        quantizer = QUANTIZERS[self.bits[state_idx]][self.quantizers[state_idx]]
        if quantizer is not None \
            and local_p.numel() >= self.min_quantizable_tensor_size \
            and local_p.numel() % packable_size == 0:
            out = quantizer.zeros(
                shape=p.shape,
                signed=signed,
                block_size=block_size,
                device=p.device,
                quantile=quantile
            )
        else:
            out = torch.zeros_like(local_p)

        # wrap subclass in DTensor as needed
        # NOTE: local tensor may have different shapes across ranks.
        # this happens when the 1st dim is not divisible by WORLD_SIZE.
        # thus, we must supply shape (and stride) to DTensor.from_local()
        if isinstance(p, DTensor):
            out = DTensor.from_local(
                local_tensor=out,
                device_mesh=p.device_mesh,
                placements=p.placements,
                run_check=False,
                shape=p.shape,
                stride=p.stride(),
            )

        return out

    @torch.no_grad()
    def step(self, closure=None):
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue

                grad = p.grad
                if grad.is_sparse:
                    raise RuntimeError("Sparse gradient is not supported")

                state = self.state[p]
                # State initialization
                if len(state) == 0:
                    state["step"] = torch.tensor(0.0)
                    state["exp_avg"] = self._init_state(p, True, group['quantile'])
                    state["exp_avg_sq"] = self._init_state(p, False, group['quantile'])

                state["step"] += 1
                step = state["step"].item()

                p_f32 = p.float()
                grad_f32 = grad.float()
                
                # weight decay
                p_f32 = p_f32.mul(1 - group['lr'] * group['weight_decay'])

                exp_avg = state['exp_avg']
                exp_avg_sq = state['exp_avg_sq']

                beta1, beta2 = group['betas']
                bias_correction1 = 1 - beta1 ** step
                bias_correction2_sqrt = math.sqrt(1 - beta2 ** step)

                # keep high precision copy for param update
                exp_avg_f32 = exp_avg.float().lerp(grad_f32, 1 - beta1)
                exp_avg_sq_f32 = exp_avg_sq.float().lerp(grad_f32.square(), 1 - beta2)
                denom = exp_avg_sq_f32.sqrt().div(bias_correction2_sqrt).add(group['eps'])

                exp_avg.copy_(exp_avg_f32)
                exp_avg_sq.copy_(exp_avg_sq_f32)

                step_size = group['lr'] / bias_correction1
                p_f32 = p_f32.addcdiv(exp_avg_f32, denom, value=-step_size)

                p.copy_(p_f32)

        return loss




import torch

from torch import Tensor


def scale_tensor(input: Tensor, block_size: int, quantile: float, eps: float = 1.e-12):
    shape = input.shape

    input = input.view(-1, block_size)
    scale = input.amax(-1).clamp_min(eps)
    alpha = input.quantile(quantile, -1) / scale
    input = input / scale.view(-1, 1)
    return input.view(shape), scale, alpha

def quantize_with_alpha(input: Tensor, alpha: Tensor, bits: int = 4):
    shape = input.shape

    logalpha = alpha.log2().view(-1, 1)
    input = input.view(logalpha.size(0), -1)
    codes = input.log2().div(logalpha).add(
        torch.rand(input.size(), generator=adaq_generator, device=input.device) - 0.5
    ).round().clip(0, 2 ** bits - 1)
    return codes.view(shape).to(torch.uint8)

def dequant_with_alpha(codes: Tensor, alpha: Tensor, scale: Tensor):
    out = (alpha.view(-1, 1) ** codes.view(alpha.size(0), -1)).mul(scale.view(-1, 1))
    return out.view(codes.shape)

def quantize_2bit_with_qmap(input: Tensor, qmap: Tensor, stochastic_rounding: bool = False):
    # GPU-friendly binary search
    # https://blog.demofox.org/2017/06/20/simd-gpu-friendly-branchless-binary-search/
    codes = torch.where(input >= qmap[2], 2, 0)
    codes += torch.where(input >= qmap[codes + 1], 1, 0)

    # rounding
    codes = codes.clip(max=3)
    codes_up = (codes + 1).clip(max=3)
    val_down = qmap[codes]
    val_up = qmap[codes_up]
    residual = input - val_down

    p = val_up - val_down
    if stochastic_rounding:
        p = p.mul(torch.rand(residual.size(), generator=adaq_generator, device=input.device))
    else:
        p = p * 0.5
    codes = torch.where(residual >= p, codes_up, codes)

    return codes.to(torch.uint8)

def quantize_3bit_with_qmap(input: Tensor, qmap: Tensor, stochastic_rounding: bool = False):
    # GPU-friendly binary search
    # https://blog.demofox.org/2017/06/20/simd-gpu-friendly-branchless-binary-search/
    codes = torch.where(input >= qmap[8], 8, 0)
    codes += torch.where(input >= qmap[codes + 4], 4, 0)
    codes += torch.where(input >= qmap[codes + 2], 2, 0)
    codes += torch.where(input >= qmap[codes + 1], 1, 0)

    # rounding
    codes = codes.clip(max=7)
    codes_up = (codes + 1).clip(max=7)
    val_down = qmap[codes]
    val_up = qmap[codes_up]
    residual = input - val_down

    p = val_up - val_down
    if stochastic_rounding:
        p = p.mul(torch.rand(residual.size(), generator=adaq_generator, device=input.device))
    else:
        p = p * 0.5
    codes = torch.where(residual >= p, codes_up, codes)

    return codes.to(torch.uint8)

def quantize_4bit_with_qmap(input: Tensor, qmap: Tensor, stochastic_rounding: bool = False):
    # GPU-friendly binary search
    # https://blog.demofox.org/2017/06/20/simd-gpu-friendly-branchless-binary-search/
    codes = torch.where(input >= qmap[8], 8, 0)
    codes += torch.where(input >= qmap[codes + 4], 4, 0)
    codes += torch.where(input >= qmap[codes + 2], 2, 0)
    codes += torch.where(input >= qmap[codes + 1], 1, 0)

    # rounding
    codes = codes.clip(max=15)
    codes_up = (codes + 1).clip(max=15)
    val_down = qmap[codes]
    val_up = qmap[codes_up]
    residual = input - val_down

    p = val_up - val_down
    if stochastic_rounding:
        p = p.mul(torch.rand(residual.size(), generator=adaq_generator, device=input.device))
    else:
        p = p * 0.5
    codes = torch.where(residual >= p, codes_up, codes)

    return codes.to(torch.uint8)

def quantize_8bit_with_qmap(input: Tensor, qmap: Tensor, stochastic_rounding: bool = False):
    # GPU-friendly binary search
    # https://blog.demofox.org/2017/06/20/simd-gpu-friendly-branchless-binary-search/
    codes = torch.where(input >= qmap[128], 128, 0)
    codes += torch.where(input >= qmap[codes + 64], 64, 0)
    codes += torch.where(input >= qmap[codes + 32], 32, 0)
    codes += torch.where(input >= qmap[codes + 16], 16, 0)
    codes += torch.where(input >= qmap[codes + 8], 8, 0)
    codes += torch.where(input >= qmap[codes + 4], 4, 0)
    codes += torch.where(input >= qmap[codes + 2], 2, 0)
    codes += torch.where(input >= qmap[codes + 1], 1, 0)

    # rounding
    codes_up = (codes + 1).clip(max=255)
    val_down = qmap[codes]
    val_up = qmap[codes_up]
    residual = input - val_down

    p = val_up - val_down
    if stochastic_rounding:
        p = p.mul(torch.rand(residual.size(), generator=adaq_generator, device=input.device))
    else:
        p = p * 0.5
    codes = torch.where(residual >= p, codes_up, codes)

    return codes.to(torch.uint8)

def init_adaq_generator():
    global adaq_generator

    if torch.distributed.is_initialized():
        seed = torch.randint(1 << 31, size=[], device=torch.device('cuda'))
        torch.distributed.broadcast(seed, src=0)
    else:
        seed = torch.randint(1 << 31, size=[])
    rank = torch.distributed.get_rank() if torch.distributed.is_initialized() else 0
    adaq_generator = torch.Generator(device=rank)
    adaq_generator.manual_seed(seed.item())

def set_block_size(n_elems: int, block_size: int) -> int:
    block_size = n_elems if block_size <= 0 else block_size
    block_size = n_elems if n_elems <= block_size else block_size
    for b in range(block_size, 0, -1):
        if n_elems % b == 0:
            return b
    return 1




from typing import Optional

import math

import torch
from torch import Tensor
from torch.utils._python_dispatch import return_and_correct_aliasing

from torchao.utils import (
    TORCH_VERSION_AT_LEAST_2_4,
    TORCH_VERSION_AT_LEAST_2_5,
    TorchAOBaseTensor,
)

from .utils import scale_tensor, quantize_with_alpha, dequant_with_alpha, set_block_size

aten = torch.ops.aten
c10d_functional = torch.ops.c10d_functional
_c10d_functional = torch.ops._c10d_functional

BITS = 4
NUMS = 2 ** BITS - 1

class QemaOptimState4bit(TorchAOBaseTensor):
    tensor_attrs = ["codes", "scale", "alpha"]

    @staticmethod
    def __new__(cls, codes: Tensor, scale: Tensor, alpha: Tensor, signed: bool, shape, quantile):
        return Tensor._make_wrapper_subclass(cls, shape, device=codes.device)

    def __init__(self, codes: Tensor, scale: Tensor, alpha: Tensor, signed: bool, shape, quantile):
        """Create quantized 4-bit optimizer state as proposed in https://arxiv.org/abs/2309.01507

        Args
            codes: quantized and packed 4-bit data stored as uint8.
            scale: scale data for block-wise quantization.
            qmap: lookup table that maps between quantized value (code) and float value.
            signed: whether the tensor is signed or unsigned.
            shape: shape of original float tensor.

        NOTE: To get block-wise scale, the original float tensor is first reshape to (-1, block_size).
        Thus, the last dimension of the original float tensor is not necessarily divisible by block size.
        Given `codes` and `scale`, `block_size` is calculated as `codes.numel() * 2 // scale.numel()`.
        The extra `* 2` is because `codes` is 4-bit data packed in 8-bit storage.
        """
        assert codes.dtype is torch.uint8
        assert codes.ndim == 1  # flattened buffer
        assert scale.ndim == 1
        assert alpha.ndim == 1
        self.codes = codes
        self.scale = scale
        self.signed = signed
        self.alpha = alpha
        self._shape = shape
        self.block_size = codes.numel() * 2 // scale.numel()
        self.quantile = quantile

    def __tensor_flatten__(self):
        return self.tensor_attrs, [self.signed, self._shape]

    @classmethod
    def __tensor_unflatten__(
        cls, tensor_data_dict, tensor_attributes, outer_size=None, outer_stride=None
    ):
        return cls(
            *[tensor_data_dict[name] for name in cls.tensor_attrs], *tensor_attributes
        )

    def dequantize(self, output_dtype=None):
        codes = torch.stack([self.codes >> 4, self.codes & 0b1111], dim=-1)  # unpack
        float_data = dequant_with_alpha(codes, self.alpha, self.scale)
        if output_dtype is not None:
            float_data = float_data.to(output_dtype)
        return float_data.view(self._shape)

    @classmethod
    def zeros(cls, shape, signed: bool = False, block_size: int = 128, quantile: float = 1.e-14, device=None, **other_kwargs):
        assert not signed, "Qema supports unsigned quantization only ..."
        shape = (shape,) if isinstance(shape, int) else shape
        n_elems = math.prod(shape)
        block_size = set_block_size(n_elems, block_size)

        codes = torch.zeros(n_elems // 2, dtype=torch.uint8, device=device)
        scale = torch.zeros(n_elems // block_size, device=device)
        alpha = torch.ones(n_elems // block_size, device=device)
        return cls(codes, scale, alpha, signed, shape, quantile)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(signed={self.signed}, block_size={self.block_size}, quantile={self.quantile}"
            f"shape={tuple(self.shape)}, device={self.device}, requires_grad={self.requires_grad})"
        )


# in pre-2.4, calling .to(device, dtype) will not dispatch aten._to_copy.default when
# dtype is the same but device is different. thus, we must override .to() method instead.
if not TORCH_VERSION_AT_LEAST_2_4:

    def _to(self, *args, **kwargs):
        # ignore other args/kwargs
        device = kwargs.pop("device", None)
        return QemaOptimState4bit(
            self.codes.to(device),
            self.scale.to(device),
            self.alpha.to(device),
            self.signed,
            self.shape,
            self.quantile,
        )

    QemaOptimState4bit.to = _to
    del _to  # make sure to not re-use


@QemaOptimState4bit.implements(aten.copy_.default)
def _(func, types, args, kwargs):
    dst = args[0]
    src = args[1]

    if isinstance(dst, QemaOptimState4bit) and isinstance(src, QemaOptimState4bit):
        assert (
            dst.signed == src.signed
            and dst.block_size == src.block_size
            and dst._shape == src._shape
        )
        dst.codes.copy_(src.codes)
        dst.scale.copy_(src.scale)
        dst.alpha.copy_(src.alpha)
    elif isinstance(dst, QemaOptimState4bit):
        scaled_src, scale, alpha = scale_tensor(src.view(-1), dst.block_size, dst.quantile)
        dst.alpha.copy_(
            alpha.pow(1 / NUMS)
        )
        codes = quantize_with_alpha(scaled_src, dst.alpha, BITS)
        dst.codes.copy_((codes[::2] << 4) | codes[1::2])  # packing
        dst.scale.copy_(scale)
    else:
        dst.copy_(src.dequantize())

    return dst


@QemaOptimState4bit.implements(aten._to_copy.default)
def _(func, types, args, kwargs):
    # ignore dtype
    device = kwargs.get("device", None)
    out = QemaOptimState4bit(
        args[0].codes.to(device=device),
        args[0].scale.to(device=device),
        args[0].alpha.to(device=device),
        args[0].signed,
        args[0].shape,
        args[0].quantile
    )
    return return_and_correct_aliasing(func, args, kwargs, out)


@QemaOptimState4bit.implements(aten.lerp.Scalar)
def _(func, types, args, kwargs):
    args = [x.dequantize() if isinstance(x, QemaOptimState4bit) else x for x in args]
    return func(*args, **kwargs)


# this is needed for DTensor.from_local() and for flattening tensor
@QemaOptimState4bit.implements(aten.view.default)
def _(func, types, args, kwargs):
    x, shape = args

    if tuple(x.shape) == tuple(shape):
        return QemaOptimState4bit(x.codes, x.scale, x.alpha, x.signed, x._shape, x.quantile)

    if len(shape) == 1 and shape[0] == -1:
        return QemaOptimState4bit(x.codes, x.scale, x.alpha, x.signed, (x.numel(),), x.quantile)

    raise ValueError(
        f"{x.__class__.__name__} only supports .view() with same shape or shape=[-1]"
    )


@QemaOptimState4bit.implements(
    [
        # required by DTensor.full_tensor()
        c10d_functional.all_gather_into_tensor.default,
        _c10d_functional.all_gather_into_tensor.default,
        c10d_functional.wait_tensor.default,
        _c10d_functional.wait_tensor.default,
        # required by torch.distributed.checkpoint.save
        aten.detach.default,
    ]
)
def _(func, types, args, kwargs):
    x = args[0]
    if not isinstance(x, QemaOptimState4bit):
        raise ValueError(f"expecting a QemaOptimState4bit but found {type(x)}")

    codes = func(x.codes, *args[1:], **kwargs)
    scale = func(x.scale, *args[1:], **kwargs)
    alpha = func(x.alpha, *args[1:], **kwargs)

    # adjust the first dim
    shape = (x._shape[0] * codes.numel() // x.codes.numel(),) + x._shape[1:]

    # assume tensors from all ranks have the same signedness
    return QemaOptimState4bit(codes, scale, alpha, x.signed, shape, x.quantile)


# required by torch.distributed.checkpoint.save
# note that we don't actually implement pin memory for this tensor subclass
# (pin_memory argument is ignored in aten._to_copy)
@QemaOptimState4bit.implements(aten.is_pinned.default)
def _(func, types, args, kwargs):
    return (
        args[0].codes.is_pinned()
        and args[0].scale.is_pinned()
        and args[0].alpha.is_pinned()
    )


# required by torch.distributed.checkpoint.load when world size changes i.e. re-sharding
@QemaOptimState4bit.implements(aten.slice.Tensor)
def _(func, types, args, kwargs):
    x, dim, start, end = args[:4]
    step = args[4] if len(args) > 4 else 1

    # input validation
    if dim != 0:
        raise ValueError("Only support aten.slice along the first dim")
    if step != 1:
        raise ValueError("Only support aten.slice with step=1")

    block_size = x.block_size
    stride = math.prod(x.shape[1:])

    # for 1 increment in x along the first dim,
    # (flattened) scale will increment by stride / block_size
    if (start * stride) % block_size != 0 or (end * stride) % block_size != 0:
        raise ValueError(
            f"Invalid start or end for shape={x.shape} and block_size={block_size}. "
            f"Make sure start and end align with block boundary. "
            f"Received start={start}, end={end}."
        )

    # note that for 4-bit, we store .codes as flattened buffer
    # divide by 2 since we store 2x 4-bit in 1x uint8
    codes = x.codes[start * stride // 2 : end * stride // 2]
    scale = x.scale[start * stride // block_size : end * stride // block_size]
    alpha = x.alpha[start * stride // block_size : end * stride // block_size]

    # adjust the first dim
    shape = (x.shape[0] * codes.numel() // x.codes.numel(),) + x.shape[1:]

    return QemaOptimState4bit(codes, scale, alpha, x.signed, shape, x.quantile)


if TORCH_VERSION_AT_LEAST_2_5:
    from torch.serialization import add_safe_globals

    add_safe_globals([QemaOptimState4bit])





import math
from typing import Any, Dict, Tuple

import torch
from torch import Tensor
from torch.distributed._tensor import DTensor
from .adamw import Optimizer, QUANTIZERS, PACKABLE, init_adaq_generator

Eps2 = Tuple[float, float]
ParamGroup = Dict[str, Any]


class AdafactorQ(Optimizer):

    r"""
    Adafactor with Quantized states.

    Parameters:
    -----------
    bits: Tuple[int, int]
        (bits for 1st state, bits for 2nd state)
    quantile: float
        quantile for 2nd state logarithmic quantization (qema)
    block_sizes: Tuple[int, int]
        (block size for 1st state, block size for 2nd state)
        Then the block-wise quantization will be performed.
        See (Dettmers T., et al. 8-bit optimizers via block-wise quantization. ICLR, 2022.)[http://arxiv.org/abs/2110.02861] for details.
    quantizers: Tuple[str, str]
        (quantizer for 1st state, quantizer for 2nd state)
        -`none`: The orginal 32-bit state
        -`bf16`: The BF16 format
        -`de`: The dynamic exponent mapping without a stochastic rounding
        -`de-sr`: The dynamic exponent mapping with a stochastic rounding
        -`linear`: The linear mapping without a stochastic rounding
        -`linear-sr`: The linear mapping with a stochastic rounding
        -`qema`: The proposed logarithmic quantization
    min_quantizable_tensor_size: int
        A tensor whose size is less than `min_quantizable_tensor_size` will be excluded from quantization.

    Examples:
    ---------
    >>> model: torch.nn.Module
    >>> from solo.adafactor import AdafactorQ
    >>> optimizer = AdafactorQ(
        model.parameters(),
        lr = 0.001,
        bits = (4, 2),
        quantile = 0.1,
        block_sizes = (128, 128),
        quantizers = ('de', 'qema'),
        min_quantizable_tensor_size = 128
    )
    """

    def __init__(
        self,
        params,
        lr = None,
        eps2: Eps2 = (1e-30, 1e-3),
        clip_threshold: float = 1.0,
        decay_rate: float = -0.8,
        beta1 = None,
        weight_decay: float = 0.0,
        scale_parameter: bool = True,
        relative_step: bool = True,
        warmup_init: bool = False,
        *,
        bits: Tuple[int] = (4, 2),
        quantile: float = 0.1,
        block_sizes: Tuple[int] = (128, 128),
        quantizers: Tuple[str] = ('de', 'qema'),
        min_quantizable_tensor_size: int = 0,
    ):
        if lr is not None and lr <= 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if weight_decay < 0.0:
            raise ValueError(
                "Invalid weight_decay value: {}".format(weight_decay)
            )

        defaults = dict(
            lr=lr,
            eps2=eps2,
            clip_threshold=clip_threshold,
            decay_rate=decay_rate,
            beta1=beta1,
            weight_decay=weight_decay,
            scale_parameter=scale_parameter,
            relative_step=relative_step,
            warmup_init=warmup_init,
            quantile=quantile
        )
        super(AdafactorQ, self).__init__(params, defaults)

        self.bits = bits
        self.quantizers = quantizers
        self.block_sizes = (int(block_sizes[0]), int(block_sizes[1]))
        self.packable_sizes = (PACKABLE[bits[0]], PACKABLE[bits[1]])
        self.min_quantizable_tensor_size = min_quantizable_tensor_size

        init_adaq_generator()

        print(self)

    def __str__(self):
        return f"AdamfactorQ: [{self.quantizers[0]}-{self.quantizers[1]}]" \
                f"[{self.bits[0]}-{self.bits[1]}]" \
                f"[{self.block_sizes[0]}-{self.block_sizes[1]}]"

    def _get_lr(self, param_group: ParamGroup, param_state) -> float:
        rel_step_sz = param_group["lr"]
        if param_group["relative_step"]:
            min_step = (
                1e-6 * param_state["step"]
                if param_group["warmup_init"]
                else 1e-2
            )
            rel_step_sz = min(min_step, 1.0 / math.sqrt(param_state["step"]))
        param_scale = 1.0
        if param_group["scale_parameter"]:
            param_scale = max(param_group["eps2"][1], param_state["RMS"])
        return param_scale * rel_step_sz

    def _get_options(
        self, param_group: ParamGroup, param_shape: Tuple[int, ...]
    ) -> Tuple[bool, bool]:
        factored = len(param_shape) >= 2
        use_first_moment = param_group["beta1"] is not None
        return factored, use_first_moment

    def _rms(self, tensor: torch.Tensor) -> float:
        return tensor.norm(2) / (tensor.numel() ** 0.5)

    def _approx_sq_grad(
        self,
        exp_avg_sq_row: torch.Tensor,
        exp_avg_sq_col: torch.Tensor,
        output: torch.Tensor,
    ) -> None:
        r_factor = (
            (exp_avg_sq_row / exp_avg_sq_row.mean(dim=-1))
            .rsqrt_()
            .unsqueeze(-1)
        )
        c_factor = exp_avg_sq_col.unsqueeze(-2).rsqrt()
        torch.mul(r_factor, c_factor, out=output)

    def _init_state(self, p: Tensor, signed: bool, quantile: float):
        local_p = p.to_local() if isinstance(p, DTensor) else p

        # follow bitsandbytes, only quantize tensors >= 4096 values
        state_idx = 0 if signed else 1
        block_size = self.block_sizes[state_idx]
        packable_size = self.packable_sizes[state_idx]
        quantizer = QUANTIZERS[self.bits[state_idx]][self.quantizers[state_idx]]
        if quantizer is not None \
            and local_p.numel() >= self.min_quantizable_tensor_size \
            and local_p.numel() % packable_size == 0:
            out = quantizer.zeros(
                shape=p.shape,
                signed=signed,
                block_size=block_size,
                device=p.device,
                quantile=quantile
            )
        else:
            out = torch.zeros_like(local_p)

        # wrap subclass in DTensor as needed
        # NOTE: local tensor may have different shapes across ranks.
        # this happens when the 1st dim is not divisible by WORLD_SIZE.
        # thus, we must supply shape (and stride) to DTensor.from_local()
        if isinstance(p, DTensor):
            out = DTensor.from_local(
                local_tensor=out,
                device_mesh=p.device_mesh,
                placements=p.placements,
                run_check=False,
                shape=p.shape,
                stride=p.stride(),
            )

        return out

    def step(self, closure = None):
        r"""Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError(
                        "Adafactor does not support sparse gradients."
                    )

                state = self.state[p]
                grad_shape = grad.shape

                factored, use_first_moment = self._get_options(
                    group, grad_shape
                )
                # State Initialization
                if len(state) == 0:
                    state["step"] = 0

                    if use_first_moment:
                        # Exponential moving average of gradient values
                        state["exp_avg"] = self._init_state(
                            grad, True, group['quantile']
                        )
                    if factored:
                        state["exp_avg_sq_row"] = self._init_state(
                            torch.zeros(grad_shape[:-1], dtype=torch.float, device=grad.device),
                            False, group['quantile']
                        )
                        state["exp_avg_sq_col"] = self._init_state(
                            torch.zeros(grad_shape[:-2] + grad_shape[-1:], dtype=torch.float, device=grad.device),
                            False, group['quantile']
                        )
                    else:
                        state["exp_avg_sq"] = self._init_state(
                            grad,
                            False, group['quantile']
                        )

                    state["RMS"] = 0

                state["step"] += 1
                state["RMS"] = self._rms(p.data)
                lr = self._get_lr(group, state)

                beta2t = 1.0 - math.pow(state["step"], group["decay_rate"])
                update = (grad**2) + group["eps2"][0]

                if factored:
                    exp_avg_sq_row = state["exp_avg_sq_row"]
                    exp_avg_sq_col = state["exp_avg_sq_col"]

                    exp_avg_sq_row_f32 = exp_avg_sq_row.float().lerp(
                        update.mean(dim=-1), 1. - beta2t
                    )
                    exp_avg_sq_col_f32 = exp_avg_sq_col.float().lerp(
                        update.mean(dim=-2), 1. - beta2t
                    )

                    # Approximation of exponential moving average of square
                    # of gradient
                    self._approx_sq_grad(
                        exp_avg_sq_row_f32, exp_avg_sq_col_f32, update
                    )

                    state["exp_avg_sq_row"].copy_(exp_avg_sq_row_f32)
                    state["exp_avg_sq_col"].copy_(exp_avg_sq_col_f32)

                    update.mul_(grad)
                else:
                    exp_avg_sq = state["exp_avg_sq"]
                    exp_avg_sq_f32 = exp_avg_sq.float().lerp(
                        update, 1. - beta2t
                    )
                    torch.rsqrt(exp_avg_sq_f32, out=update).mul_(grad)

                    state["exp_avg_sq"].copy_(exp_avg_sq_f32)

                update.div_(
                    max(1.0, self._rms(update) / group["clip_threshold"])
                )
                update.mul_(lr)

                if use_first_moment:
                    exp_avg = state["exp_avg"]
                    exp_avg.mul_(group["beta1"]).add_(
                        update, alpha=1 - group["beta1"]
                    )
                    update = exp_avg

                if group["weight_decay"] != 0:
                    p.data.add_(p.data, alpha=-group["weight_decay"] * lr)

                p.data.add_(-update)

        return loss




from typing import Tuple

import torch, math

from .adamw import AdamWQ


class AdaBeliefQ(AdamWQ):

    def __str__(self):
        return f"AdaBeliefQ: [{self.quantizers[0]}-{self.quantizers[1]}]" \
                f"[{self.bits[0]}-{self.bits[1]}]" \
                f"[{self.block_sizes[0]}-{self.block_sizes[1]}]"

    @torch.no_grad()
    def step(self, closure=None):
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue

                grad = p.grad
                if grad.is_sparse:
                    raise RuntimeError("Sparse gradient is not supported")

                state = self.state[p]
                # State initialization
                if len(state) == 0:
                    state["step"] = torch.tensor(0.0)
                    state["exp_avg"] = self._init_state(p, True, group['quantile'])
                    state["exp_avg_sq"] = self._init_state(p, False, group['quantile'])

                state["step"] += 1
                step = state["step"].item()

                p_f32 = p.float()
                grad_f32 = grad.float()
                
                # weight decay
                p_f32 = p_f32.mul(1 - group['lr'] * group['weight_decay'])

                exp_avg = state['exp_avg']
                exp_avg_sq = state['exp_avg_sq']

                beta1, beta2 = group['betas']
                bias_correction1 = 1 - beta1 ** step
                bias_correction2_sqrt = math.sqrt(1 - beta2 ** step)

                # keep high precision copy for param update
                exp_avg_f32 = exp_avg.float().lerp(grad_f32, 1 - beta1)
                grad_residual = grad_f32 - exp_avg_f32
                exp_avg_sq_f32 = exp_avg_sq.float().lerp(grad_residual.square(), 1 - beta2)
                denom = exp_avg_sq_f32.sqrt().div(bias_correction2_sqrt).add(group['eps'])

                exp_avg.copy_(exp_avg_f32)
                exp_avg_sq.copy_(exp_avg_sq_f32)

                step_size = group['lr'] / bias_correction1
                p_f32 = p_f32.addcdiv(exp_avg_f32, denom, value=-step_size)

                p.copy_(p_f32)

        return loss





from .subclass_2bit import FixedLinearOptimState2bit, FixedDEOptimState2bit
from .subclass_3bit import FixedLinearOptimState3bit, FixedDEOptimState3bit
from .subclass_4bit import FixedLinearOptimState4bit, FixedDEOptimState4bit
from .subclass_8bit import FixedLinearOptimState8bit, FixedDEOptimState8bit

from .de_2bit import DEOptimState2bit
from .de_3bit import DEOptimState3bit
from .de_4bit import DEOptimState4bit
from .de_8bit import DEOptimState8bit

from .linear_2bit import LinearOptimState2bit
from .linear_3bit import LinearOptimState3bit
from .linear_4bit import LinearOptimState4bit
from .linear_8bit import LinearOptimState8bit

from .de_sr_2bit import DESROptimState2bit
from .de_sr_3bit import DESROptimState3bit
from .de_sr_4bit import DESROptimState4bit
from .de_sr_8bit import DESROptimState8bit

from .linear_sr_2bit import LinearSROptimState2bit
from .linear_sr_3bit import LinearSROptimState3bit
from .linear_sr_4bit import LinearSROptimState4bit
from .linear_sr_8bit import LinearSROptimState8bit

from .qema_2bit import QemaOptimState2bit
from .qema_3bit import QemaOptimState3bit
from .qema_4bit import QemaOptimState4bit
from .qema_8bit import QemaOptimState8bit

from .bf16 import BF16




import torch, math

from torchao.prototype.low_bit_optim.subclass_8bit import OptimState8bit
from torchao.prototype.low_bit_optim.quant_utils import create_dynamic_map
from .utils import set_block_size


QMAP_SIGNED_LINEAR = torch.linspace(-1, 1, 256).tolist()
QMAP_UNSIGNED_LINEAR = torch.linspace(0, 1, 257)[1:].tolist()

QMAP_SIGNED_DE = create_dynamic_map(signed=True, max_exponent_bits=7, total_bits=8)
QMAP_UNSIGNED_DE = create_dynamic_map(signed=False, max_exponent_bits=7, total_bits=8)


class FixedLinearOptimState8bit(OptimState8bit):

    @classmethod
    def zeros(cls, shape, signed: bool = True, block_size: int = 256, device=None, **other_kwargs):
        shape = (shape,) if isinstance(shape, int) else shape
        n_elems = math.prod(shape)
        block_size = set_block_size(n_elems, block_size)

        codes = torch.zeros(shape, dtype=torch.uint8, device=device)
        scale = torch.zeros(codes.numel() // block_size, device=device)
        qmap = torch.tensor(QMAP_SIGNED_LINEAR if signed else QMAP_UNSIGNED_LINEAR, device=device)
        return cls(codes, scale, qmap, signed)


class FixedDEOptimState8bit(OptimState8bit):

    @classmethod
    def zeros(cls, shape, signed: bool = True, block_size: int = 256, device=None, **other_kwargs):
        shape = (shape,) if isinstance(shape, int) else shape
        n_elems = math.prod(shape)
        block_size = set_block_size(n_elems, block_size)

        codes = torch.zeros(shape, dtype=torch.uint8, device=device)
        scale = torch.zeros(codes.numel() // block_size, device=device)
        qmap = torch.tensor(QMAP_SIGNED_DE if signed else QMAP_UNSIGNED_DE, device=device)
        return cls(codes, scale, qmap, signed)




import torch


class BF16:

    @classmethod
    def zeros(cls, shape, signed: bool = True, block_size: int = 0, device=None, **other_kwargs):
        return torch.zeros(shape, dtype=torch.bfloat16, device=device)


import setuptools
import re

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("LICENSE", "r") as f:
    license_text = f.read()

requires = [
    "torchao>=0.7.0",
]

def get_property(prop, project):
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), open("src/" + project + '/__init__.py').read())
    return result.group(1)

setuptools.setup(
    name="solow",
    packages=setuptools.find_packages(where="src"),
    version=get_property('__version__', 'solo'),
    package_dir={"solo": "src/solo"},
    author="MTandHJ",
    author_email="congxueric@gmail.com",
    description="SOLO",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT',
    license_files=('LICENSE',),
    python_requires='>=3.8',
    install_requires=requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
)




import math

import torch
from torch import Tensor
from torch.utils._python_dispatch import return_and_correct_aliasing

from torchao.utils import (
    TORCH_VERSION_AT_LEAST_2_4,
    TORCH_VERSION_AT_LEAST_2_5,
    TorchAOBaseTensor,
)

from torchao.prototype.low_bit_optim.quant_utils import (
    dequant_with_qmap,
    scale_tensor,
)
from .utils import set_block_size, quantize_4bit_with_qmap

aten = torch.ops.aten
c10d_functional = torch.ops.c10d_functional
_c10d_functional = torch.ops._c10d_functional

QMAP_SIGNED_LINEAR = torch.linspace(-1, 1, 16).tolist()
QMAP_UNSIGNED_LINEAR = torch.linspace(0, 1, 17)[1:].tolist()

class LinearSROptimState4bit(TorchAOBaseTensor):
    tensor_attrs = ["codes", "scale", "qmap"]

    @staticmethod
    def __new__(cls, codes: Tensor, scale: Tensor, qmap: Tensor, signed: bool, shape):
        return Tensor._make_wrapper_subclass(cls, shape, device=codes.device)

    def __init__(self, codes: Tensor, scale: Tensor, qmap: Tensor, signed: bool, shape):
        """Create quantized 4-bit optimizer state as proposed in https://arxiv.org/abs/2309.01507

        Args
            codes: quantized and packed 4-bit data stored as uint8.
            scale: scale data for block-wise quantization.
            qmap: lookup table that maps between quantized value (code) and float value.
            signed: whether the tensor is signed or unsigned.
            shape: shape of original float tensor.

        NOTE: To get block-wise scale, the original float tensor is first reshape to (-1, block_size).
        Thus, the last dimension of the original float tensor is not necessarily divisible by block size.
        Given `codes` and `scale`, `block_size` is calculated as `codes.numel() * 2 // scale.numel()`.
        The extra `* 2` is because `codes` is 4-bit data packed in 8-bit storage.
        """
        assert codes.dtype is torch.uint8
        assert codes.ndim == 1  # flattened buffer
        assert scale.ndim == 1
        self.codes = codes
        self.scale = scale
        self.qmap = qmap
        self.signed = signed
        self._shape = shape
        self.block_size = codes.numel() * 2 // scale.numel()

    def __tensor_flatten__(self):
        return self.tensor_attrs, [self.signed, self._shape]

    @classmethod
    def __tensor_unflatten__(
        cls, tensor_data_dict, tensor_attributes, outer_size=None, outer_stride=None
    ):
        return cls(
            *[tensor_data_dict[name] for name in cls.tensor_attrs], *tensor_attributes
        )

    def dequantize(self, output_dtype=None):
        codes = torch.stack([self.codes >> 4, self.codes & 0b1111], dim=-1)  # unpack
        float_data = dequant_with_qmap(codes, self.qmap, self.scale)
        if output_dtype is not None:
            float_data = float_data.to(output_dtype)
        return float_data.view(self._shape)

    @classmethod
    def zeros(cls, shape, signed: bool = True, block_size: int = 0, device=None, **other_kwargs):
        shape = (shape,) if isinstance(shape, int) else shape
        n_elems = math.prod(shape)
        block_size = set_block_size(n_elems, block_size)

        codes = torch.zeros(n_elems // 2, dtype=torch.uint8, device=device)
        scale = torch.zeros(n_elems // block_size, device=device)
        qmap = torch.tensor(QMAP_SIGNED_LINEAR if signed else QMAP_UNSIGNED_LINEAR, device=device)
        return cls(codes, scale, qmap, signed, shape)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(signed={self.signed}, block_size={self.block_size}, "
            f"shape={tuple(self.shape)}, device={self.device}, requires_grad={self.requires_grad})"
        )


# in pre-2.4, calling .to(device, dtype) will not dispatch aten._to_copy.default when
# dtype is the same but device is different. thus, we must override .to() method instead.
if not TORCH_VERSION_AT_LEAST_2_4:

    def _to(self, *args, **kwargs):
        # ignore other args/kwargs
        device = kwargs.pop("device", None)
        return LinearSROptimState4bit(
            self.codes.to(device),
            self.scale.to(device),
            self.qmap.to(device),
            self.signed,
            self.shape,
        )

    LinearSROptimState4bit.to = _to
    del _to  # make sure to not re-use


@LinearSROptimState4bit.implements(aten.copy_.default)
def _(func, types, args, kwargs):
    dst = args[0]
    src = args[1]

    if isinstance(dst, LinearSROptimState4bit) and isinstance(src, LinearSROptimState4bit):
        assert (
            dst.signed == src.signed
            and dst.block_size == src.block_size
            and dst._shape == src._shape
        )
        dst.codes.copy_(src.codes)
        dst.scale.copy_(src.scale)
        # qmap should be the same, don't need to copy

    elif isinstance(dst, LinearSROptimState4bit):
        scaled_src, scale = scale_tensor(src.view(-1), dst.block_size)
        codes = quantize_4bit_with_qmap(scaled_src, dst.qmap, stochastic_rounding=True)
        dst.codes.copy_((codes[::2] << 4) | codes[1::2])  # packing
        dst.scale.copy_(scale)

    else:
        dst.copy_(src.dequantize())

    return dst


@LinearSROptimState4bit.implements(aten._to_copy.default)
def _(func, types, args, kwargs):
    # ignore dtype
    device = kwargs.get("device", None)
    out = LinearSROptimState4bit(
        args[0].codes.to(device=device),
        args[0].scale.to(device=device),
        args[0].qmap.to(device=device),
        args[0].signed,
        args[0].shape,
    )
    return return_and_correct_aliasing(func, args, kwargs, out)


@LinearSROptimState4bit.implements(aten.lerp.Scalar)
def _(func, types, args, kwargs):
    args = [x.dequantize() if isinstance(x, LinearSROptimState4bit) else x for x in args]
    return func(*args, **kwargs)


# this is needed for DTensor.from_local() and for flattening tensor
@LinearSROptimState4bit.implements(aten.view.default)
def _(func, types, args, kwargs):
    x, shape = args

    if tuple(x.shape) == tuple(shape):
        return LinearSROptimState4bit(x.codes, x.scale, x.qmap, x.signed, x._shape)

    if len(shape) == 1 and shape[0] == -1:
        return LinearSROptimState4bit(x.codes, x.scale, x.qmap, x.signed, (x.numel(),))

    raise ValueError(
        f"{x.__class__.__name__} only supports .view() with same shape or shape=[-1]"
    )


@LinearSROptimState4bit.implements(
    [
        # required by DTensor.full_tensor()
        c10d_functional.all_gather_into_tensor.default,
        _c10d_functional.all_gather_into_tensor.default,
        c10d_functional.wait_tensor.default,
        _c10d_functional.wait_tensor.default,
        # required by torch.distributed.checkpoint.save
        aten.detach.default,
    ]
)
def _(func, types, args, kwargs):
    x = args[0]
    if not isinstance(x, LinearSROptimState4bit):
        raise ValueError(f"expecting a LinearSROptimState4bit but found {type(x)}")

    codes = func(x.codes, *args[1:], **kwargs)
    scale = func(x.scale, *args[1:], **kwargs)

    # adjust the first dim
    shape = (x._shape[0] * codes.numel() // x.codes.numel(),) + x._shape[1:]

    # assume tensors from all ranks have the same signedness
    return LinearSROptimState4bit(codes, scale, x.qmap.clone(), x.signed, shape)


# required by torch.distributed.checkpoint.save
# note that we don't actually implement pin memory for this tensor subclass
# (pin_memory argument is ignored in aten._to_copy)
@LinearSROptimState4bit.implements(aten.is_pinned.default)
def _(func, types, args, kwargs):
    return (
        args[0].codes.is_pinned()
        and args[0].scale.is_pinned()
        and args[0].qmap.is_pinned()
    )


# required by torch.distributed.checkpoint.load when world size changes i.e. re-sharding
@LinearSROptimState4bit.implements(aten.slice.Tensor)
def _(func, types, args, kwargs):
    x, dim, start, end = args[:4]
    step = args[4] if len(args) > 4 else 1

    # input validation
    if dim != 0:
        raise ValueError("Only support aten.slice along the first dim")
    if step != 1:
        raise ValueError("Only support aten.slice with step=1")

    block_size = x.block_size
    stride = math.prod(x.shape[1:])

    # for 1 increment in x along the first dim,
    # (flattened) scale will increment by stride / block_size
    if (start * stride) % block_size != 0 or (end * stride) % block_size != 0:
        raise ValueError(
            f"Invalid start or end for shape={x.shape} and block_size={block_size}. "
            f"Make sure start and end align with block boundary. "
            f"Received start={start}, end={end}."
        )

    # note that for 4-bit, we store .codes as flattened buffer
    # divide by 2 since we store 2x 4-bit in 1x uint8
    codes = x.codes[start * stride // 2 : end * stride // 2]
    scale = x.scale[start * stride // block_size : end * stride // block_size]

    # adjust the first dim
    shape = (x.shape[0] * codes.numel() // x.codes.numel(),) + x.shape[1:]

    return LinearSROptimState4bit(codes, scale, x.qmap.clone(), x.signed, shape)


if TORCH_VERSION_AT_LEAST_2_5:
    from torch.serialization import add_safe_globals

    add_safe_globals([LinearSROptimState4bit])
